#define ZASM_PROCESSOR_C 1

#include <zasm/processor.h>
#include <zasm/memory.h>
#include <zasm/string.h>
#include <zasm/lexer.h>
#include <zasm/feeder.h>

#include <string.h>

#define IS_EXPR_TOKEN(T) (((T) == TOK_SPACE)||((T) == TOK_NAME)||((T) == TOK_DOT)||((T) == TOK_INTEGER)||\
                         ((T) == TOK_OP_IFTHEN)||((T) == TOK_OP_ELSE)||\
                         ((T) == TOK_OP_LOG_OR)||((T) == TOK_OP_LOG_AND)||((T) == TOK_OP_LOG_XOR)||((T) == TOK_OP_LOG_NOT)||\
                         ((T) == TOK_OP_BIN_OR)||((T) == TOK_OP_BIN_AND)||((T) == TOK_OP_BIN_XOR)||((T) == TOK_OP_BIN_NOT)||\
                         ((T) == TOK_OP_SHIFT_LEFT)||((T) == TOK_OP_SHIFT_RIGHT)||((T) == TOK_OP_EQU)||((T) == TOK_OP_NEQ)||\
                         ((T) == TOK_OP_LT)||((T) == TOK_OP_GT)||((T) == TOK_OP_LE)||((T) == TOK_OP_GE)||((T) == TOK_OP_CMP)||\
                         ((T) == TOK_OP_ADD)||((T) == TOK_OP_SUB)||((T) == TOK_OP_MUL)||((T) == TOK_OP_DIV)||\
                         ((T) == TOK_OP_MOD)||((T) == TOK_OP_POW))

struct processor_t *processor_new()
{
  struct processor_t *val;
  ALLOCATE(struct processor_t, val, 1);
  processor_fill(val);
  return val;
}

void processor_fill(struct processor_t *proc)
{
  proc->endianness = PROC_LITTLE_ENDIAN;
  proc->instruction_set = NULL;
}

struct processor_t *processor_read(char *filename)
{
  struct feeder_t *input;
  struct lexer_t *lex;
  struct token_t token,result;

  lex = lexer_new(string_new(filename));
  input = feeder_new();
  feeder_add_action(input, action_new(ACTION_EXECUTE_T(lexer_execute), NULL, NULL, lex));
  token = feeder_get(input);
  result = processor_p(input, &token);
  if ((result.type != TOK_PROCESSOR) || (token.type != TOK_END_OF_FILE))
    {
      T_ERROR(token,"syntax error\n");
    }
  feeder_free(input);
  return result.value.t_struct_processor_t_p;
}

void processor_add_instructions(struct processor_t *proc, struct list_t *instrs)
{
  struct proc_instr_t *instr_new;
  struct proc_instr_t *instr_old;
  struct token_t t;
  foreach(t,instrs)
    {
      instr_new = t.value.t_struct_proc_instr_t_p;
      instr_old = processor_get_instr(proc, instr_new->name);
      if (instr_old)
	{
	  list_add_list_begin(instr_old->arguments, instr_new->arguments);
	  proc_instr_free(instr_new);
	}
      else
	list_add_end(proc->instruction_set, t);
    }
  list_free(instrs);
}

struct proc_instr_t *processor_get_instr(struct processor_t *proc, char *instr)
{
  struct token_t t;

  foreach(t,proc->instruction_set)
    if (0==strcasecmp(t.value.t_struct_proc_instr_t_p->name, instr))
      return t.value.t_struct_proc_instr_t_p;
  return NULL;
}

void processor_delete(struct processor_t *proc)
{
  list_free(proc->instruction_set);
}

void processor_free(struct processor_t *proc)
{
  processor_delete(proc);
  free(proc);
}

struct proc_instr_t *proc_instr_new()
{
  struct proc_instr_t *val;
  ALLOCATE(struct proc_instr_t, val, 1);
  proc_instr_fill(val);
  return val;
}

void proc_instr_fill(struct proc_instr_t *instr)
{
  instr->name = NULL;
  instr->arguments = NULL;
}

struct proc_instr_arg_t *proc_instr_get_arg(struct proc_instr_t *instr, struct list_t *args)
{
  struct token_t t;

  foreach(t,instr->arguments)
    if (proc_instr_arg_check_pattern(t.value.t_struct_proc_instr_arg_t_p->pattern, args))
      return t.value.t_struct_proc_instr_arg_t_p;
  return NULL;
}

void proc_instr_delete(struct proc_instr_t *instr)
{
  string_free(instr->name);
  list_free(instr->arguments);
}

void proc_instr_free(struct proc_instr_t *instr)
{
  proc_instr_delete(instr);
  free(instr);
}

struct proc_instr_arg_t *proc_instr_arg_new()
{
  struct proc_instr_arg_t *val;
  ALLOCATE(struct proc_instr_arg_t, val, 1);
  proc_instr_arg_fill(val);
  return val;
}

void proc_instr_arg_fill(struct proc_instr_arg_t *instr)
{
  instr->pattern = NULL;
  instr->data = NULL;
}

int proc_instr_arg_check_pattern(struct list_t *pat, struct list_t *args)
{
  struct token_t t,a;
  int d;

  list_start_begin(args);
  a = list_read_begin(args);
  foreach (t,pat)
    {
      if (t.type == TOK_SPACE)
	continue;
      if (t.type == TOK_OP_MUL)
	{
	  d = 0;
	  while (1)
	    {
	      if (a.type == TOK_PAREN_OPEN)
		d++;
	      else if (a.type == TOK_PAREN_CLOSE)
		{
		  d--;
		  if (d < 0)
		    break;
		}
	      else if (!IS_EXPR_TOKEN(a.type))
		break;
	      a = list_read_begin(args);
	    }
	  continue;
	}
      while (a.type == TOK_SPACE)
	a = list_read_begin(args);
      if (a.type != t.type)
	return 0;
      if (a.type == TOK_NAME)
	{
	  if (strcasecmp(a.value.t_char_p,t.value.t_char_p))
	    return 0;
	}
      else if (a.type == TOK_INTEGER)
	if (a.value.t_int != t.value.t_int)
	  return 0;
      a = list_read_begin(args);
    }
  while (a.type == TOK_SPACE)
    a = list_read_begin(args);
  if (a.type != TOK_LIST_END)
    return 0;
  return 1;
}

struct list_t *proc_instr_arg_get_pattern(struct list_t *pat, struct list_t *args)
{
  struct token_t t,a;
  struct list_t *vals, *val;
  int d;

  vals = list_new();
  list_start_begin(args);
  a = list_read_begin(args);
  foreach (t,pat)
    {
      if (t.type == TOK_SPACE)
	continue;
      if (t.type == TOK_OP_MUL)
	{
	  d = 0;
	  val = list_new();
	  while (1)
	    {
	      while (a.type == TOK_SPACE)
		a = list_read_begin(args);
	      if (a.type == TOK_PAREN_OPEN)
		d++;
	      else if (a.type == TOK_PAREN_CLOSE)
		{
		  d--;
		  if (d < 0)
		    break;
		}
	      else if (!IS_EXPR_TOKEN(a.type))
		break;
	      list_add_end(val, a);
	      a = list_read_begin(args);
	    }
	  t.type = TOK_LIST;
	  t.value.t_struct_list_t_p = val;
	  list_add_end(vals, t);
	  continue;
	}
      while (a.type == TOK_SPACE)
	a = list_read_begin(args);
      if (a.type != t.type)
	return 0;
      if (a.type == TOK_NAME)
	{
	  if (strcasecmp(a.value.t_char_p,t.value.t_char_p))
	    return NULL;
	}
      else if (a.type == TOK_INTEGER)
	if (a.value.t_int != t.value.t_int)
	  return NULL;
      a = list_read_begin(args);
    }
  while (a.type == TOK_SPACE)
    a = list_read_begin(args);
  if (a.type != TOK_LIST_END)
    return NULL;
  return vals;
}

void proc_instr_arg_delete(struct proc_instr_arg_t *instr)
{
  list_free(instr->pattern);
  list_free(instr->data);
}

void proc_instr_arg_free(struct proc_instr_arg_t *instr)
{
  proc_instr_arg_delete(instr);
  free(instr);
}
