#define ZASM_TASMTAB_C 1
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <zasm/tasmtab.h>
#include <zasm/lexer.h>
#include <zasm/processor.h>
#include <zasm/errors.h>
#include <zasm/string.h>
#include <zasm/memory.h>

typedef void (*tasm_rule_exec_t)(int nargs,struct list_t *code,int or,int length,int class,int shift);

struct tasm_rule_t
{
  char *rule;
  tasm_rule_exec_t exec;
};

#define ADD_ARG(CODE, LENGTH, FORMAT, ARGS...) \
(({\
  if (shift && or) add_arg(CODE, LENGTH, "((" FORMAT ")<<%i)|%i", ## ARGS, shift, or);\
  else if (shift) add_arg(CODE, LENGTH, "(" FORMAT ")<<%i", ## ARGS, shift);\
  else if (or) add_arg(CODE, LENGTH, "(" FORMAT ")|%i", ## ARGS, or);\
  else add_arg(CODE, LENGTH, "(" FORMAT ")", ## ARGS);\
}))

static void add_arg(struct list_t *code, int length, char *format, ...);
static int args_count(struct list_t *lst);

static void rule_nop(int nargs,struct list_t *code,int or,int length,int class,int shift);
static void rule_r1(int nargs,struct list_t *code,int or,int length,int class,int shift);
static void rule_r2(int nargs,struct list_t *code,int or,int length,int class,int shift);
static void rule_crel(int nargs,struct list_t *code,int or,int length,int class,int shift);
static void rule_swap(int nargs,struct list_t *code,int or,int length,int class,int shift);
static void rule_combine(int nargs,struct list_t *code,int or,int length,int class,int shift);
static void rule_cswap(int nargs,struct list_t *code,int or,int length,int class,int shift);
static void rule_zbit(int nargs,struct list_t *code,int or,int length,int class,int shift);
static void rule_zidx(int nargs,struct list_t *code,int or,int length,int class,int shift);

static struct tasm_rule_t rules[] =
  {
    {"NOP"    , rule_nop},
    {"NOTOUCH", rule_nop},
    {"R1"     , rule_r1},
    {"R2"     , rule_r2},
    {"CREL"   , rule_crel},
    {"SWAP"   , rule_swap},
    {"COMBINE", rule_combine},
    {"CSWAP"  , rule_cswap},
    {"ZBIT"   , rule_zbit},
    {"ZIDX"   , rule_zidx},
    {NULL, NULL}
  };

struct proc_instr_t *tasmtab_convert(char *name,
                                     struct list_t *args,
                                     struct list_t *code,
                                     int length,
                                     char *rule,
                                     int class,
                                     int shift,
                                     int or)
{
  struct proc_instr_t *instr;
  struct proc_instr_arg_t *arg;
  struct token_t t;
  struct list_t *lst;
  int i;

  for (i = 0; rules[i].rule; i++)
    if (strcasecmp(rules[i].rule, rule) == 0)
      rules[i].exec(args_count(args), code, or, length, class, shift);
  string_free(rule);

  arg = proc_instr_arg_new();
  arg->pattern = args;
  arg->data = code;
  t.type = TOK_PROC_INSTR_ARG;
  t.value.t_struct_proc_instr_arg_t_p = arg;
  lst = list_new();
  list_add_end(lst, t);
  instr = proc_instr_new();
  instr->name = name;
  instr->arguments = lst;
  return instr;
}

static void add_arg(struct list_t *code, int length, char *format, ...)
{
  char *str = NULL;
  int str_l = 16;
  int val = 0;
  va_list ap;
  struct token_t t;
  struct list_t *l;

  switch (length)
    {
    case 0:
      return;
    case 1:
      t.type = TOK_BYTE;
      break;
    case 2:
      t.type = TOK_WORD;
      break;
    case 4:
      t.type = TOK_DWORD;
      break;
    default:
      ERROR("invalid length of opcode argument: %i\n", length);
    }
  list_add_end(code, t);

  va_start(ap, format);
  do
    {
      REALLOCATE(char, str, str_l);
      val = vsnprintf(str, str_l, format, ap);
      if (val == str_l)
	str_l+=16;
      else if (val > str_l)
	str_l = val;
      else
	break;
    } while (1);
  va_end(ap);

  l = lexer_tokenize_string(str);
  list_add_list_end(code, l);
  list_free(l);
}

static int args_count(struct list_t *lst)
{
  struct token_t t;
  int n = 0;
  foreach(t,lst)
    {
      if (t.type == TOK_OP_MUL)
	n++;
    }
  return n;
}

// the rule definitions

static void rule_nop(int nargs,struct list_t *code,int or,int length,int class,int shift)
{
  if (nargs == 0)
    return;
  else if (nargs == 1)
    ADD_ARG(code, length - list_length(code), "@0");
  else
    ERROR("invalid TASM NOP rule\n");
}

static void rule_r1(int nargs,struct list_t *code,int or,int length,int class,int shift)
{
  if (nargs == 0)
    return;
  else if ((nargs == 1) && ((length - list_length(code)) == 1))
    ADD_ARG(code, 1, "(@0)-$.end");
  else
    ERROR("invalid TASM R1 rule\n");
}

static void rule_r2(int nargs,struct list_t *code,int or,int length,int class,int shift)
{
  if (nargs == 0)
    return;
  else if ((nargs == 1) && ((length - list_length(code)) == 2))
    ADD_ARG(code, 2, "(@0)-$.end");
  else
    ERROR("invalid TASM R2 rule\n");
}

static void rule_crel(int nargs,struct list_t *code,int or,int length,int class,int shift)
{
  if ((nargs == 2) && ((length - list_length(code)) == 2))
    ADD_ARG(code, 2, "((@0).&0xff).|((((@1)-$.end).&0xff)<<8)");
  else
    ERROR("invalid TASM CREL rule\n");
}

static void rule_swap(int nargs,struct list_t *code,int or,int length,int class,int shift)
{
  if ((nargs == 1) && ((length - list_length(code)) == 2))
    ADD_ARG(code, 2, "(((@0).&0xff00)>>8).|(((@0).&0xff)<<8)");
  else
    ERROR("invalid TASM SWAP rule\n");
}

static void rule_combine(int nargs,struct list_t *code,int or,int length,int class,int shift)
{
  if ((nargs == 2) && ((length - list_length(code)) == 2))
    ADD_ARG(code, 2, "((@0).&0xff).|(((@1).&0xff)<<8)");
  else
    ERROR("invalid TASM COMBINE rule\n");
}

static void rule_cswap(int nargs,struct list_t *code,int or,int length,int class,int shift)
{
  if ((nargs == 2) && ((length - list_length(code)) == 2))
    ADD_ARG(code, 2, "((@1).&0xff).|(((@0).&0xff)<<8)");
  else
    ERROR("invalid TASM CSWAP rule\n");
}

static void rule_zbit(int nargs,struct list_t *code,int or,int length,int class,int shift)
{
  int cl = list_length(code);
  struct token_t t;
  if ((nargs == 1) && (length - cl == 0))
    {
      t = list_get_end(code);
      if (t.type != TOK_BYTE_INT)
	ERROR("invalid TASM ZBIT rule\n");
      ADD_ARG(code, 1, "(((@0).&0x07)<<3).|%i", t.value.t_int);
    }
  else if (nargs == 1)
    ADD_ARG(code, length-cl, "(((@0).&0x07)<<3)<<%i", (length-cl-1)*8);
  else if (nargs == 2)
    ADD_ARG(code, length-cl, "((((@0).&0x07)<<3)<<%i).|(@1)", (length-cl-1)*8);
  else
    ERROR("invalid TASM ZBIT rule\n");  
}

static void rule_zidx(int nargs,struct list_t *code,int or,int length,int class,int shift)
{
  if (nargs == 0)
    return;
  else if (nargs == 1)
    ADD_ARG(code, length - list_length(code), "(@0).&0xff");
  else
    ERROR("invalid TASM ZIDX rule\n");  
}
